"""
run_file(workflowFilePath::String, system::OrchestratorRegistry.ServiceRegistry)

Executes a workflow defined in a file, leveraging a provided `ServiceRegistry` to manage the services involved.

This function automates the setup, execution, and teardown of a containerized workflow, simplifying the process of running complex workflows with multiple dependencies. It performs the following steps:
- Parses the workflow file specified by `workflowFilePath` to extract tasks.
- Retrieves the `FACTResultsIO` service from the provided `ServiceRegistry` to configure the results container.
- Manages workflow execution, including handling tasks, updating global variables, and interacting with services.
- Optionally updates the results container with the final workflow results.

# Arguments
- `workflowFilePath::String`: The path to the file containing the workflow definition.
- `system::OrchestratorRegistry.ServiceRegistry`: A registry of services required for workflow execution.

# Behavior
- Parses the workflow file to determine the tasks to execute.
- Uses the `FACTResultsIO` service from the registry to set up a results container.
- Executes tasks defined in the workflow, interacting with the necessary services as defined in the registry.
- Manages cleanup and updates for the results container after workflow execution.

# Side Effects
- Starts and stops Docker containers as required for workflow execution.
- Modifies the results container based on workflow outcomes.

# Example
To run a workflow defined in a file located at `/path/to/workflow.yml`, using a preconfigured service registry `system`:

> run_file("/path/to/workflow.yml", system)

This will execute the workflow, managing all service interactions and container operations as needed.
"""
function run_file(workflowFilePath::String, system::OrchestratorRegistry.ServiceRegistry)
    # Parse the workflow file
    tasks = parseWorkflow(workflowFilePath)

    # Create a reference to results container
    service  = OrchestratorRegistry.get_services_by_name(system, "FACTResultsIO")[1]
    results_container = Dict(
        "user" => "root",
        "password" => "devops",
        "host" => service.host,
        "port" => service.port
    )

    # Execute the workflow
    manage_workflow(tasks, results_container, system)
end

"""
manage_workflow(tasks::Dict, results_container::Dict, system::OrchestratorRegistry.ServiceRegistry)

Executes a workflow by managing the execution of tasks based on their dependencies and the provided service registry.

This function orchestrates the execution of tasks in a workflow, ensuring that each task waits for its dependencies to complete before starting. It uses multithreading to execute tasks in parallel while respecting dependencies and resource constraints. Each task interacts with the provided results container and may trigger updates to it.

# Arguments
- `tasks::Dict`: A dictionary where keys are task names and values are dictionaries containing task details. Task details include:
  - `dependencies`: A list of other task names that must complete before this task can run.
  - Other relevant task-specific information needed for execution.
- `results_container::Dict`: A dictionary representing the results container, used by tasks to store or retrieve results.
- `system::OrchestratorRegistry.ServiceRegistry`: The service registry used to manage services required by tasks during execution.

# Returns
- `Ref{Bool}`: A reference to a boolean flag indicating whether the results container requires saving due to updates during task execution.

# Behavior
- Initializes channels for each task to signal dependency completion.
- Spawns a thread for each task, ensuring it waits for its dependencies to signal completion.
- Uses the `system` parameter to provide access to necessary services during task execution.
- Updates the `results_container` and tracks whether it requires saving using an atomic boolean flag.

# Example
To execute a workflow, define the tasks, results container, and service registry:

```julia
tasks = Dict(
    "task1" => Dict("dependencies" => [], "info" => "example task 1"),
    "task2" => Dict("dependencies" => ["task1"], "info" => "example task 2")
)

results_container = Dict("user" => "root", "password" => "devops", "host" => "localhost", "port" => 8080)

system = ServiceRegistry()
# Add required services to the registry
add_service!(system, Service("123", "example-service", "http://localhost", 9000))

# Execute the workflow
update_needed = manage_workflow(tasks, results_container, system)
println("Results container needs saving: ", update_needed[])
```

This example demonstrates how tasks are executed respecting dependencies, with access to services provided by the registry.
"""
function manage_workflow(tasks::Dict, results_container::Dict, system::OrchestratorRegistry.ServiceRegistry)
    # Flag to return whether there is a need to save an updated image of FACTResultsIO
    update_condition = Atomic{Bool}(false)

    # Initialize a Dict to hold Channels for each task
    task_channels = Dict{String, Channel{Bool}}()

    for task_name in keys(tasks)
        # For simplicity, use an unbuffered channel
        task_channels[task_name] = Channel{Bool}(1)
    end

    task_handles = []  # Initialize an array to hold task handles

    # Start the tasks logic
    for (task_name, task_info) in tasks
        #println("Spawning task: $task_name")
        t = Threads.@spawn begin
            try
                # Prepare the dependencies information beforehand
                dependencies_info = Dict{String, Dict}()
                for dep in task_info["dependencies"]
                    dependencies_info[dep] = tasks[dep]
                end
                # Check if the task has dependencies and wait for them if it does
                if haskey(task_info, "dependencies") && !isempty(task_info["dependencies"])
                    for dep in task_info["dependencies"]
                        # Wait for the dependency task to signal completion
                        #println("[$task_name] waiting on dependency $dep")
                        take!(task_channels[dep])
                        # put back the signal into the channel because somebody else may need it!!!
                        put!(task_channels[dep],true)
                        #println("[$task_name] dependency $dep completed")
                    end
                end
                # Execute the task
                #println("Executing task: $task_name")
                run_task(task_name, task_info, dependencies_info, system, results_container, update_condition, overwrite)
                #println("Finished executing task: $task_name")
            catch ex
                println("Error occurred in task $task_name: ", ex)
                println("Stacktrace: ", stacktrace(catch_backtrace()))
            finally
                put!(task_channels[task_name], true)
                #println("Finished executing task: $task_name")
            end
        end
        push!(task_handles, t)  # Collect the task handle
    end
    
    # Wait on each task to complete
    #println("Waiting for all tasks to complete...")
    foreach(wait, task_handles)  # This waits for each task to complete

    println("All tasks completed.")
end
